function [lambdas,y,S,iters] = multisvr(kernels,labels,C,alpha,lambdasO,biasflag,epsilon)
%
% function [lambdas,y,S,iters] = multisvr(kernels,labels,C,alpha,lambdasO)
%
%   Modified version of multisvm (TJ2009) to do regression instead of
%   classification
%
% It is highly recommended to replace the quadprog function
% call with a fast SVM solver (such as SVMperf or Pegasos).
% At the very least please try use MOSEK's quadprog
% function instead of the default Matlab quadprog. 
%
% The lambdasO value is optional and is useful for warm-starts.
% Make sure that it satisfies the SVM constraints ahead of time
% (setting it zero always do this).
%
% THIS VERSION STARTS BOUND AT ZERO BUT ONLY USES LAMBDASO TO SEED QP
%
% Copyright 2009 by Tony Jebara
% Adaptation for regression by Dan Gillespie (DGillespie@cosineaudio.com) 2010

if (nargin < 7)
    epsilon = 1e-6;
end
if (nargin < 6)
    biasFlag = 1;       %this really needs to be set based on the kernel used
end

% Allocate Output Variables
M   = size(labels,2);
D   = size(kernels,2);
S   = ones(D,1);
s   = ones(D,1);
G   = ones(D,1);
Y   = kernels;
T   = zeros(M,1);
for m=1:M
  T(m) = size(labels{m},1);
    y_hat{m} = [labels{m}; -labels{m}];
end

e = epsilon;

if (nargin<5)
  lambdasO = labels;    %this just creates a lambdasO that's the same size as lambdas
  for m=1:M
    lambdasO{m} = zeros(2*T(m),1);
  end
end;


lambdas = labels;       %this just creates a lambdas that's the same size as labels
for m=1:M
  lambdas{m} = zeros(2*T(m),1);
end



% Compute Predictions and Switches Given Current Lambdas
for d=1:D
  s(d) = 0;
  for m=1:M
    Y{m,d} = [kernels{m,d} -kernels{m,d}]*lambdas{m};      %y_hat_m_t(d)
    s(d)   = s(d) + lambdas{m}'*[kernels{m,d} -kernels{m,d}; -kernels{m,d} kernels{m,d}]*lambdas{m};        %is this true?
  end
  if (s(d)<(-1000))     %bound s(d) to avoid exp() error
    s(d)=-1000; 
  end
  G(d) = alpha*exp(-0.5*s(d));      %g(d)
  if (G(d)>0)                       %what if g(d) is <= 0?
    G(d) = tanh(log(G(d))/2)/(2*log(G(d)));     %G(d)
  end
  S(d) = 1/(1+alpha*exp(-0.5*s(d)));            %s_hat(d)
end

% Compute the objective function        USE THE CODE FROM THE QP SETUP FOR
% THIS      %ACTUALLY THE QP MIGHT BE ABOUT CALCULATING THE BOUND NOT THE j
J = 0;
for m=1:M
	J = J - e*lambdas{m}'*ones(2*T(m),1) + lambdas{m}'*y_hat{m};      
end
for d=1:D
	J = J - log(alpha + exp(0.5*s(d)));   % the format of this should be right, but the way it's calced above may be the same but better.  Also, it may include the Dlog(alpha+1) term which I'm dropping
end         %I guess we don't calculate the D*log(alpha + 1) part because it's a constant - or maybe it's tied into that odd commented out J calculation
% fprintf('Iteration=0 J=%e\n',J);

going = 1;
iters = 1;
while (going)

  old = lambdas;        %old = lambda_tilda
  qpchange = 0;
  for m=1:M
    % Solve the individual QP's
    vlB  = zeros(2*T(m),1);           %s.t. 0<=lambda_m_t<=C        This has to be for all lambda_hat
    vuB  = C*ones(2*T(m),1);          %s.t. 0<=lambda_m_t<=C
    if biasFlag == 1
        B = [];, c = [];
    else
        B = [ones(1,T(m)) -ones(1,T(m))];, c = 0;     % Set the constraint Ax = b     %This assumes we're calculating bias explicitly (or that there isn't one)%
    end

    A = [];, b = [];
    
    grad = -y_hat{m} + e*ones(2*T(m),1);  %this is instead of grad-hess*lambdas{m} - I don't know if this is right now

    Hd = zeros(2*T(m),2*T(m));
    hess = zeros(2*T(m),2*T(m));
    for d=1:D
      Hd = [kernels{m,d} -kernels{m,d}; -kernels{m,d} kernels{m,d}];
	  hess = hess + G(d)*Hd*lambdas{m}*lambdas{m}'*Hd + Hd;
	  grad = grad + S(d)*Hd*lambdas{m};
    end

%    hess = hess + eye(2*T(m))/(2*C);            %regularization

    valOLD = (grad-hess*old{m})'*lambdas{m}+0.5*lambdas{m}'*hess*lambdas{m};      
    old{m} = lambdas{m};


    lambdas{m} = quadprog(hess,grad-hess*lambdas{m},A,b,B,c,vlB,vuB,lambdasO{m});     %remove from classification problem
%    lambdas{m} = quadprog(Hb,c,[],[],A,b,vlB,vuB,lambdasO{m});
    lambdasO{m} = lambdas{m};
    valNEW = (grad-hess*old{m})'*lambdas{m}+0.5*lambdas{m}'*hess*lambdas{m};
    qpchange = qpchange+valOLD-valNEW;

    % Interlave update for Y, S and G after each model changes...
    for d=1:D
    Y{m,d} = [kernels{m,d} -kernels{m,d}]*lambdas{m};      %y_hat_m_t(d)
    s(d)   = s(d) + lambdas{m}'*[kernels{m,d} -kernels{m,d}; -kernels{m,d} kernels{m,d}]*lambdas{m};
    s(d)   = s(d) - old{m}'*[kernels{m,d} -kernels{m,d}; -kernels{m,d} kernels{m,d}]*old{m};            %This used to be backward but I changed it
      if (s(d)<(-1000)) 
          s(d)=-1000; 
      end;
      G(d)   = alpha*exp(-0.5*s(d));
      if (G(d)>0)
        G(d) = tanh(log(G(d))/2)/(2*log(G(d)));
      end
      S(d) = 1/(1+alpha*exp(-0.5*s(d)));
    end
  end

  % To Be Safe, Cleanly Compute Predictions and Switches Given Current Lambdas
  for d=1:D
    s(d) = 0;
    for m=1:M
      Y{m,d} = [kernels{m,d} -kernels{m,d}]*lambdas{m};      %y_hat_m_t(d)
      s(d)   = s(d) + lambdas{m}'*[kernels{m,d} -kernels{m,d}; -kernels{m,d} kernels{m,d}]*lambdas{m};
    end
    if (s(d)<(-1000)) 
      s(d)=-1000; 
    end
    G(d) = alpha*exp(-0.5*s(d));
    if (G(d)>0)
      G(d) = tanh(log(G(d))/2)/(2*log(G(d)));
    end
    S(d) = 1/(1+alpha*exp(-0.5*s(d)));
  end


  % Compute the objective function
  J = 0;
  for m=1:M
	J = J - e*lambdas{m}'*ones(2*T(m),1) + lambdas{m}'*y_hat{m};      
  end
  for d=1:D
	J = J - log(alpha + exp(0.5*s(d)));   % the format of this should be right, but the way it's calced above may be the same but better.  Also, it may include the Dlog(alpha+1) term which I'm dropping
  end
  
  % Check if lambdas converged
  delt = 0;
  for m=1:M
    delt = max(delt,max(abs(lambdas{m}-old{m})));
%    plot(abs(lambdas{m}-old{m}))           %temp test
%    pause;                     
  end
  %[iters delt] %temp test show progress
  if (delt<(C*1e-3))
    going = 0;
  end

  % Check if number of iterations was exceeded
%  if (iters>1000)
  if (iters>200)        %this is taking too long
    going = 0;
    fprintf('Exceeded max iterations stoping early.\n');
  end
  iters = iters+1;
end

% Finally label the training data
y = labels;
for m=1:M
  y{m} = zeros(T(m),1);
  for d=1:D
    y{m} = y{m}+S(d)*Y{m,d};
  end
end

fprintf('C=%e alpha=%e Iteration=%d J=%e qpimprove=%e\n',C,alpha,iters,J,qpchange);

